import numpy as np
from loguru import logger

from illico.utils.groups import GroupContainer
from illico.utils.registry import data_handler_registry


def log_memory_usage(data_handler, grpc: GroupContainer, batch_size: int, n_threads: int) -> None:
    """Log estimated memory usage of the whole routine.

    This function simply computes estimated memory footprint of the DE genes routine.

    Parameters
    ----------
    data_handler : DataHandler
    grpc : GroupContainer
        GroupContainer
    batch_size : int
        Batch size used for processing genes
    n_threads : int
        Number of threads used for processing genes

    Author: Rémy Dubois
    """
    X = data_handler.data
    data_handler = data_handler_registry.get(X)
    # Results hold fold change, p-values, statistics
    results_fp = grpc.counts.size * X.shape[1] * 3 * np.float64().nbytes
    # Instead of just doing batch_size*n_cells, we scale it by the total object size so that sparsity is taken into account,
    # and scale it by number of threads running concurrently
    proc_fp = batch_size / X.shape[1] * n_threads * data_handler.footprint()
    total_fp = results_fp + proc_fp
    logger.trace(
        f"Estimated RAM footprint: {total_fp / (1000**3):.2f} GB. (This includes {results_fp / (1000**3):.3f} GB for results and {proc_fp / (1000**3):.3f} GB for processing data.)"
    )
    return total_fp



import anndata as ad
import memray
import pytest

from illico import asymptotic_wilcoxon
from illico.utils.compile import _precompile
from illico.utils.groups import encode_and_count_groups
from illico.utils.memory import log_memory_usage
from illico.utils.registry import data_handler_registry


@pytest.mark.skip(
    f"Memory footprint is actually too cumbersome to estimate, especially in very low data regimes like in this test."
)
@pytest.mark.parametrize("test", ["ovo", "ovr"])
def test_log_memory_usage(rand_adata, test, tmp_path):
    # No need to test that exception is raised, as it is done in `test_asymptotic_wilcoxon` already
    if isinstance(rand_adata.X, ad._core.sparse_dataset._CSRDataset):
        pytest.skip("CSR lazy data not supported for now.")

    data_handler = data_handler_registry.get(rand_adata.X)
    reference = rand_adata.obs.pert.iloc[0] if test == "ovo" else None
    _, grpc = encode_and_count_groups(groups=rand_adata.obs["pert"].tolist(), ref_group=reference)

    batch_size = 16
    n_threads = 1

    estimated_heap_size = log_memory_usage(data_handler, grpc, batch_size, n_threads)

    # precompile
    _precompile(data_handler, reference)

    with memray.Tracker((_f := tmp_path / "tracker.bin")) as tracker:
        asymptotic_wilcoxon(
            adata=rand_adata,
            is_log1p=False,
            group_keys="pert",
            reference=reference,
            n_threads=n_threads,
            batch_size=batch_size,
        )
    with memray.FileReader(_f) as reader:
        real_heap_size = reader.metadata.peak_memory
    # Allow some tolerance for overhead
    assert real_heap_size <= estimated_heap_size * 1.1, f"Estimated: {estimated_heap_size}, Real: {real_heap_size}"



from collections import namedtuple
from typing import Any

import numpy as np

GroupContainer = namedtuple(
    "GroupContainer",
    [
        "encoded_groups",
        "counts",
        "indices",
        "indptr",
        "encoded_ref_group",
    ],
)


def encode_and_count_groups(groups: np.ndarray, ref_group: Any) -> tuple[np.ndarray, GroupContainer]:
    """Build the GroupContainer holding all group-related information.

    GroupContainer holds:
    - original group information
    - reference group (control)
    - encoded groups
    - unique raw groups
    - counts (of cell, per group)
    - indices, indptr in a RLE format
    - encoded reference group (control)

    Args:
        groups (np.ndarray): 1-d array holding group labels, one per cell
        ref_group (Any): Flag

    Returns:
        unique_groups (np.ndarray): Array of unique group labels
        GroupContainer: GroupContainer holding all group-related information.

    Author: Rémy Dubois
    """
    if ref_group not in groups and ref_group is not None:
        raise ValueError(f"Reference group `{ref_group}` is not present in the group labels.")
    # Determine group indices
    group_indices = np.argsort(groups, kind="stable").astype(np.uint64)

    # Count occcurrences of each group
    sorted_groups = groups[group_indices]
    change_idx = np.flatnonzero(sorted_groups[1:] != sorted_groups[:-1]) + 1
    group_counts = np.diff(np.r_[0, change_idx, sorted_groups.size]).astype(np.uint64)

    # Find unique groups
    unique_groups = sorted_groups[np.r_[0, change_idx]]

    # Encode groups as integers
    encoded_groups = np.searchsorted(unique_groups, groups).astype(np.uint64)

    # Build indptr
    group_indptr = np.cumsum(np.insert(group_counts, 0, 0)).astype(np.uint64)

    return unique_groups, GroupContainer(
        encoded_groups=encoded_groups,
        counts=group_counts.astype(np.uint64),
        indices=group_indices,
        indptr=group_indptr,
        encoded_ref_group=(
            -1 if ref_group is None else int(np.searchsorted(unique_groups, ref_group))
        ),  # Weirdly enough, this must be -1 and not None, otherwise Numba fails to compile various functions, especially branching
    )



import os
import sys

# Add project root to sys.path so autodoc can import the package
sys.path.insert(0, os.path.abspath(".."))

project = "illico"
author = "illico contributors"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.napoleon",
    "sphinx.ext.viewcode",
    "myst_parser",
]

# Allow parsing Markdown files with MyST
myst_enable_extensions = [
    "dollarmath",
    "amsmath",
]

autosummary_generate = True

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]


# Prefer the Furo theme when available, fall back otherwise
html_theme = "furo"

html_static_path = ["_static"]

# Mock imports that may not be installed in the build environment so autodoc can
# import the package even when heavy scientific deps are missing.
autodoc_mock_imports = [
    "anndata",
    "joblib",
    "loguru",
    "tqdm",
    "numba",
    "h5py",
    "numpy",
    "scipy",
    "pandas",
    "sklearn",
]

# Autodoc settings
autodoc_member_order = "bysource"
autodoc_typehints = "description"



from illico.ovr.dense_ovr import dense_ovr_mwu_kernel_over_contiguous_col_chunk
from illico.ovr.sparse_ovr import (
    csc_ovr_mwu_kernel_over_contiguous_col_chunk,
    csr_ovr_mwu_kernel_over_contiguous_col_chunk,
)

__all__ = [
    "dense_ovr_mwu_kernel_over_contiguous_col_chunk",
    "csc_ovr_mwu_kernel_over_contiguous_col_chunk",
    "csr_ovr_mwu_kernel_over_contiguous_col_chunk",
]



import os
import urllib.request
from pathlib import Path
from tempfile import NamedTemporaryFile
from typing import Literal

import anndata as ad
import numpy as np
import pandas as pd
import pytest
from scipy import sparse
from tqdm import tqdm

from illico.utils.registry import data_handler_registry

CACHE_ROOT = Path(os.environ.get("ILLICO_PYTEST_CACHE", "/tmp/pytest_cache"))
CELL_LINE_URLS = {
    "k562": "https://plus.figshare.com/ndownloader/files/35773219",
    "rpe1": "https://plus.figshare.com/ndownloader/files/35775606",
    "jurkat": "https://www.ncbi.nlm.nih.gov/geo/download/?acc=GSE264667&format=file&file=GSE264667%5Fjurkat%5Fraw%5Fsinglecell%5F01%2Eh5ad",
    "hepg2": "https://www.ncbi.nlm.nih.gov/geo/download/?acc=GSE264667&format=file&file=GSE264667%5Fhepg2%5Fraw%5Fsinglecell%5F01%2Eh5ad",
}


@pytest.fixture(
    params=[
        (cell_line, fmt, fraction)
        for cell_line in ["k562", "rpe1", "jurkat", "hepg2"]
        for fmt in ["dense", "csr", "csc"]
        for fraction in [0.0, 0.2, 1.0]
    ],
    scope="function",
    ids=lambda p: f"{p[0]}-{p[1]}-{p[2]:.0%}",
)
def adata(request):
    """Fixture to download, convert and cache cell line dataset with subsampling"""
    cell_line, fmt, fraction = request.param
    # if fraction.values[0] < 1.0:
    #     request.node.add_marker("debug")

    cache_root = CACHE_ROOT / cell_line
    cache_root.mkdir(parents=True, exist_ok=True)
    raw_path = cache_root / f"{cell_line}.h5ad"
    download_if_missing(cell_line, raw_path)

    target_path = cache_root / f"{cell_line}_{fmt}_{fraction:.0%}.h5ad"

    if target_path.exists():
        adata = ad.read_h5ad(target_path)
        if fmt == "dense":
            adata.X += 1.0
            adata.X -= 1.0
    else:
        if fmt == "csr":
            adata = ad.read_h5ad(raw_path, as_sparse="X", as_sparse_fmt=sparse.csr_matrix)
        elif fmt == "csc":
            adata = ad.read_h5ad(raw_path, as_sparse="X", as_sparse_fmt=sparse.csc_matrix)
        else:
            adata = ad.read_h5ad(raw_path)

        if fraction == 0.0:
            col_idxs = np.random.RandomState(0).choice(adata.n_vars, size=1, replace=False)
            adata = adata[:, col_idxs].copy()
        elif fraction < 1.0:
            col_idxs = np.random.RandomState(0).choice(adata.n_vars, size=int(adata.n_vars * fraction), replace=False)
            # Sort the column indices because that's how scipy builds sparse matrices.
            col_idxs.sort()
            adata = adata[:, col_idxs].copy()

        adata.write_h5ad(target_path)

    return adata


# TODO: params on log1p and normalization ? A lot of tests would result
@pytest.fixture(
    scope="function",
    params=[(fmt, lazy) for fmt in ["dense", "csc", "csr"] for lazy in [False, True]],
    ids=lambda p: f"{p[0]}-{'lazy' if p[1] else 'eager'}",
)
def rand_adata(request, tmp_path):
    n_cells = 10_000
    n_genes = 15
    n_groups = 5
    assert n_groups >= 2
    sparsity = 0.5  # ~50% zeros
    rng = np.random.RandomState(0)

    # Random gene-specific mean expression levels
    gene_means = rng.uniform(0.1, 15, size=n_genes)

    # Sample Poisson counts
    dense_counts = rng.poisson(gene_means, size=(n_cells, n_genes)).astype(
        np.float64
    )  # scipy 1.17+ returns the same dtype as input

    # Impose ~50% sparsity by random masking
    mask = rng.rand(n_cells, n_genes) < sparsity
    dense_counts[mask] = 0

    # Create groups associated
    groups = rng.randint(0, n_groups, size=n_cells)  # Add one ref group

    fmt, lazy = request.param
    if fmt == "dense":
        data_matrix = dense_counts
    elif fmt == "csc":
        data_matrix = sparse.csc_matrix(dense_counts)
    elif fmt == "csr":
        data_matrix = sparse.csr_matrix(dense_counts)
    else:
        raise ValueError(f"Unknown data format: {fmt}")

    adata = ad.AnnData(
        data_matrix,
        obs=pd.DataFrame({"pert": [f"pert_{g}" for g in groups]}),
        var=pd.DataFrame(index=[f"gene_{i}" for i in range(n_genes)]),
    )
    if lazy:
        adata_path = tmp_path / f"rand_adata_{fmt}_lazy.h5ad"
        adata.write_h5ad(adata_path)
        adata = ad.read_h5ad(adata_path, backed="r")
    return adata


@pytest.fixture(scope="function")
def eager_rand_adata(rand_adata):
    if rand_adata.isbacked:
        pytest.skip("This fixture returns only in-RAM dataset.")
    return rand_adata


def download_if_missing(cell_line: Literal["k562", "rpe1"], dst: Path) -> Path:
    """
    Stream-download k562-essential. Skips if exists.
    """
    url = CELL_LINE_URLS[cell_line]
    chunk_size = 10 * 1024 * 1024
    dst.parent.mkdir(parents=True, exist_ok=True)

    if dst.exists():
        return dst

    req = urllib.request.Request(
        url,
        headers={"User-Agent": "Mozilla/5.0"},
    )

    with urllib.request.urlopen(req) as resp:
        total = resp.length  # may be None, but Figshare provides it
        with NamedTemporaryFile(delete=False, dir=CACHE_ROOT) as tmp_file:
            try:
                with (
                    open(tmp_file.name, "wb") as f,
                    tqdm(
                        total=total,
                        unit="B",
                        unit_scale=True,
                        unit_divisor=1024,
                        smoothing=0.1,
                        desc=f"Downloading {dst}",
                    ) as pbar,
                ):
                    while True:
                        chunk = resp.read(chunk_size)
                        if not chunk:
                            break
                        f.write(chunk)
                        pbar.update(len(chunk))
            except Exception as e:
                # Clean up temp file on error
                Path(tmp_file.name).unlink(missing_ok=True)
                raise e
            # Move temp file to final destination
            Path(tmp_file.name).rename(dst)

    return dst


@pytest.fixture(scope="function")
def rand_csr():
    n_rows = 1000
    n_cols = 200
    density = 0.1
    rng = np.random.RandomState(0)
    data = sparse.random(n_rows, n_cols, density=density, format="csr", random_state=rng, dtype=np.float64)
    handler = data_handler_registry.get(data)
    return handler.to_nb(data)


@pytest.fixture(scope="function")
def rand_csc():
    n_rows = 1000
    n_cols = 200
    density = 0.1
    rng = np.random.RandomState(0)
    data = sparse.random(n_rows, n_cols, density=density, format="csc", random_state=rng, dtype=np.float64)
    handler = data_handler_registry.get(data)
    return handler.to_nb(data)


# def limit_memory_if_lazy(memory_limit: str):

#     def decorator(func):
#         def wrapper(*args, **kwargs):
#             data_handler: DataHandler = data_handler_registry.get(args[0])
#             if data_handler.is_lazy(args[0]):
#                 data_handler.set_memory_limit(args[0], memory_limit)
#             return func(*args, **kwargs)

#         return wrapper

#     return decorator



import contextlib
import gc
import os
import re
import warnings
from pathlib import Path

import anndata as ad
import memray
import numpy as np
import pandas as pd
import pytest
import scanpy as sc
from numba import set_num_threads
from pdex import parallel_differential_expression
from pdex._single_cell import parallel_differential_expression_vec_wrapper
from scipy import sparse as py_sparse
from scipy.stats import mannwhitneyu

from illico.asymptotic_wilcoxon import asymptotic_wilcoxon
from illico.utils.compile import _precompile
from illico.utils.registry import data_handler_registry, nb_dispatcher_registry

set_num_threads(1)  # Ensure single-threaded by default for testing consistency

ATOL = 0.0
RTOL = 1.0e-12


def scanpy_mannwhitneyu(adata, groupby_key, reference):
    sc.tl.rank_genes_groups(
        adata,
        groupby=groupby_key,
        method="wilcoxon",
        use_raw=False,
        tie_correct=True,
        reference=reference or "rest",
    )

    rg = adata.uns["rank_genes_groups"]
    groups = rg["names"].dtype.names  # perturbed genes

    records = []

    for g in groups:
        records.append(
            pd.DataFrame(
                {
                    "target": g,
                    "feature": rg["names"][g],
                    "pval": rg["pvals"][g],
                    "pval_adj": rg["pvals_adj"][g],
                    "logfoldchange": rg["logfoldchanges"][g],
                    "ustat": rg["scores"][g],  # Wilcoxon statistic proxy
                }
            )
        )

    df = pd.concat(records, ignore_index=True)
    return df.set_index(["target", "feature"])


def scipy_mannwhitneyu(adata, groupby_key, reference, use_continuity, alternative, is_log1p=False):
    if reference is not None:
        ref_counts = adata[adata.obs[groupby_key].eq(reference)].X
        if not isinstance(ref_counts, np.ndarray):
            ref_counts = ref_counts.toarray()

    # Loop over perturbations
    results = []
    for pert in adata.obs[groupby_key].unique():
        if pert == reference:
            continue
        mask = adata.obs[groupby_key].eq(pert).values
        grp_counts = adata.X[mask]  # Grab the perturbed counts
        if reference is None:
            ref_counts = adata.X[~mask]  # Grab the perturbed counts

        # densify
        if not isinstance(grp_counts, np.ndarray):
            grp_counts = grp_counts.toarray()
        if not isinstance(ref_counts, np.ndarray):
            ref_counts = ref_counts.toarray()

        # Compute FC
        if is_log1p:
            grp_counts = np.expm1(grp_counts)
            ref_counts = np.expm1(ref_counts)
            fc = np.expm1(grp_counts).mean(axis=0) / np.expm1(ref_counts).mean(axis=0)
        else:
            fc = np.mean(grp_counts, axis=0) / np.mean(ref_counts, axis=0)

        stats, pvals = mannwhitneyu(
            ref_counts, grp_counts, axis=0, method="asymptotic", use_continuity=use_continuity, alternative=alternative
        )
        results.append(
            pd.DataFrame(
                {
                    "p_value": pvals,
                    "statistic": stats,
                    "fold_change": fc,
                    "target": pert,
                    "feature": adata.var_names,
                }
            )
        )
    results = pd.concat(results, axis=0).set_index(["target", "feature"])
    return results


@pytest.mark.parametrize("use_rust", [True, False], ids=["rust", "numba"])
@pytest.mark.parametrize("alternative", ["two-sided", "less", "greater"])
@pytest.mark.parametrize("tie_correct", [True, False], ids=["tie-correct", "no-tie-correct"])
@pytest.mark.parametrize("use_continuity", [True, False], ids=["contin-corr", "no-contin-corr"])
@pytest.mark.parametrize("test", ["ovo", "ovr"])
def test_asymptotic_wilcoxon(rand_adata, test, use_continuity, tie_correct, alternative, use_rust):
    if not rand_adata.isbacked:
        cached = rand_adata.copy()

    if test == "ovo":
        reference = rand_adata.obs.pert.iloc[0]
    else:
        reference = None

    # If rand_adata is lazy and CSR, ensure we raise an error because this is not supported
    if isinstance(rand_adata.X, ad._core.sparse_dataset._CSRDataset):
        ctx = pytest.raises(
            KeyError,
            match="Support for data type <class 'anndata._core.sparse_dataset._CSRDataset'> is not implemented.",
        )
        should_raise = True
    else:
        ctx = contextlib.nullcontext()
        should_raise = False

    with ctx:
        asy_results = asymptotic_wilcoxon(
            adata=rand_adata,
            is_log1p=False,
            group_keys="pert",
            reference=reference,
            use_continuity=use_continuity,
            tie_correct=tie_correct,
            n_threads=1,
            batch_size=16,
            alternative=alternative,
            use_rust=use_rust,
        )

    if should_raise:
        return

    if not tie_correct:
        # We skip at this point, so that we make sure that at least the code runs
        pytest.skip(f"Skipping comparison with scipy when tie correction is disabled, as scipy does not support it.")

    scipy_results = scipy_mannwhitneyu(
        adata=rand_adata,
        groupby_key="pert",
        reference=reference,
        is_log1p=False,
        use_continuity=use_continuity,
        alternative=alternative,
    )
    # sc_results = scanpy_mannwhitneyu(adata=rand_adata, groupby_key="pert", reference=reference)

    np.testing.assert_allclose(
        asy_results.loc[scipy_results.index].statistic.values,
        scipy_results.statistic.values,
        atol=0.0,
        rtol=0.0,
    )
    # idxs = np.where(np.isclose(asy_results.loc[scipy_results.index].statistic.values, scipy_results.statistic.values, rtol=0., atol=1.0e-12))[0]
    # x = asy_results.loc[scipy_results.index].p_value.values[idxs]
    # y = scipy_results.p_value.values[idxs]
    # import ipdb; ipdb.set_trace()
    # Test statistics exactly
    # Test p-values with low tolerance
    np.testing.assert_allclose(
        asy_results.loc[scipy_results.index].p_value.values,
        scipy_results.p_value.values,
        atol=0.0,
        rtol=1.0e-12,
    )
    # Test FC with mid tolerance
    np.testing.assert_allclose(
        asy_results.loc[scipy_results.index].fold_change.values,
        scipy_results.fold_change.values,
        atol=0.0,
        rtol=1.0e-6,
    )

    if not rand_adata.isbacked:
        # Test that the original adata is not modified, some sorting happen in-place so just making sure
        pd.testing.assert_frame_equal(rand_adata.obs, cached.obs)
        pd.testing.assert_frame_equal(rand_adata.var, cached.var)
        if isinstance(rand_adata.X, np.ndarray):
            np.testing.assert_array_equal(rand_adata.X, cached.X)
        else:
            np.testing.assert_array_equal(rand_adata.X.toarray(), cached.X.toarray())


# Do not sweep all the possible test params, alternative and all
@pytest.mark.parametrize("use_rust", [True, False], ids=["rust", "numba"])
@pytest.mark.parametrize("backed", [True, False], ids=["lazy", "eager"])
@pytest.mark.parametrize("test", ["ovo", "ovr"])
def test_backed_asymptotic_wilcoxon(eager_rand_adata, test, backed, use_rust, tmp_path):
    # No need to test that exception is raised, as it is done in `test_asymptotic_wilcoxon` already
    if isinstance(eager_rand_adata.X, py_sparse.csr.csr_matrix) and backed:
        pytest.skip("CSR lazy data not supported for now.")

    if test == "ovo":
        reference = eager_rand_adata.obs.pert.iloc[0]
    else:
        reference = None

    # Precompile
    data_handler = data_handler_registry.get(eager_rand_adata.X)
    _precompile(data_handler, reference)

    # Run this with one thread and small batch size, this simply makes sure we never load
    adata_path = tmp_path / f"rand_adata_lazy.h5ad"
    # Make this anndata bigger, otherwise memory measurements are not significant
    bigger_eager_rand_adata = ad.concat([eager_rand_adata] * 300, axis=1)
    # Concatenation converts to CSR, so revert back to CSC
    if isinstance(eager_rand_adata.X, py_sparse.csc.csc_matrix):
        bigger_eager_rand_adata.X = py_sparse.csc_matrix(bigger_eager_rand_adata.X)
    bigger_eager_rand_adata.obs = eager_rand_adata.obs.copy()
    bigger_eager_rand_adata.var_names_make_unique()
    bigger_eager_rand_adata.write_h5ad(adata_path)

    # In order to track proper memory usage, we include the read_h5ad call within the memray context
    # Consequently, memory allocated to adata will show as heap memory, unlike memory tests below which only
    # tracked algorithm allocations
    with memray.Tracker(tmp_path / "memray-trace.bin", file_format=memray.FileFormat.AGGREGATED_ALLOCATIONS):
        adata = ad.read_h5ad(adata_path, backed="r" if backed else None)
        _ = asymptotic_wilcoxon(
            adata=adata,
            is_log1p=False,
            group_keys="pert",
            reference=reference,
            use_continuity=True,
            n_threads=1,
            batch_size=16,
            alternative="two-sided",
            use_rust=use_rust,
        )
    max_rss, max_heap = 0, 0
    with memray.FileReader(tmp_path / "memray-trace.bin") as reader:
        for snapshot in reader.get_memory_snapshots():
            max_rss = max(max_rss, snapshot.rss)
            max_heap = max(max_heap, snapshot.heap)
    print(f"Max RSS: {max_rss/1_000_000:.1f} MB, Max heap: {max_heap/1_000_000:.1f} MB")
    if backed:
        if max_heap > 30_000_000:  # 30 MB
            raise AssertionError(
                f"Expected low (<30MB) heap memory usage when running in backed mode, got {max_heap/1_000_000:.1f} MB."
            )
    else:
        if max_heap < 200_000_000:  # 200 MB
            raise AssertionError(
                f"Expected high (>200MB) heap memory usage when running in eager mode, got {max_heap/1_000_000:.1f} MB."
            )


def test_unsorted_indices_error(eager_rand_adata):
    """Test that an error is raised if input data matrix indices are not sorted."""
    if isinstance(eager_rand_adata.X, np.ndarray):
        pytest.skip("Test only relevant for sparse matrices.")
    # Unsort the indices of the csr matrix
    eager_rand_adata.X.indices[:] = eager_rand_adata.X.indices[::-1]
    with pytest.raises(ValueError):
        _ = asymptotic_wilcoxon(
            adata=eager_rand_adata,
            is_log1p=False,
            group_keys="pert",
            reference="non-targeting",
            n_threads=1,
            batch_size=16,
        )


def call_routine(data, method, test, num_threads, use_rust):
    def run():
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            if method == "pdex":
                parallel_differential_expression(
                    data,
                    groupby_key="gene",
                    reference="non-targeting",
                    num_workers=num_threads,
                )
            elif method == "pdexp":
                parallel_differential_expression_vec_wrapper(
                    data,
                    groupby_key="gene",
                    reference="non-targeting",
                    num_workers=num_threads,
                )
            elif method == "illico":
                reference = "non-targeting" if test == "ovo" else None
                asymptotic_wilcoxon(
                    data,
                    is_log1p=False,
                    group_keys="gene",
                    reference=reference,
                    n_threads=num_threads,
                    batch_size="auto",
                    use_rust=use_rust,
                )
            elif method == "scanpy":
                reference = "non-targeting" if test == "ovo" else "rest"
                set_num_threads(num_threads)  # Scanpy does not set number of threads explicitely
                group_counts = data.obs["gene"].value_counts()
                valid_groups = group_counts.index[group_counts.values > 1].tolist()
                sc.tl.rank_genes_groups(
                    data,
                    groupby="gene",
                    groups=valid_groups,
                    reference=reference,
                    method="wilcoxon",
                    tie_correct=True,
                )
            else:
                raise ValueError(method)

    return run


@pytest.mark.speed_bench
@pytest.mark.parametrize("use_rust", [True, False], ids=["rust", "numba"])
@pytest.mark.parametrize("num_threads", [1, 2, 4, 8, 16], ids=lambda v: f"nthreads={v}")
@pytest.mark.parametrize("test", ["ovo", "ovr"])
@pytest.mark.parametrize("method", ["illico", "scanpy", "pdex", "pdexp"])
def test_speed_benchmark(adata, method, test, num_threads, use_rust, benchmark, request):
    """Not a test, just a speed benchmark."""
    if test != "ovo" and method in ["pdex", "pdexp"]:
        # This exits the test, not running the benchmark, and not raising an error
        pytest.skip("pdex only implements OVO test.")

    # Compile
    if method == "illico":
        _precompile(data_handler_registry.get(adata.X), reference="non-targeting" if test == "ovo" else None)

    params = re.match(".*\[(.*)\]", request.node.name).group(1).split("-")
    group_params = [p for i, p in enumerate(params) if i in [0, 1, 4]]
    benchmark.group = "-".join(group_params)
    _ = benchmark.pedantic(
        call_routine(adata, method, test, num_threads, use_rust), iterations=1, warmup_rounds=0, rounds=1
    )


@pytest.mark.memory_bench
@pytest.mark.parametrize("num_threads", [1, 8], ids=lambda v: f"nthreads={v}")
@pytest.mark.parametrize("test", ["ovo", "ovr"])
@pytest.mark.parametrize("method", ["illico", "scanpy", "pdex", "pdexp"])
def test_memory_benchmark(adata, method, test, num_threads, request):
    """Not a test, just a memory footprint benchmark."""
    if test != "ovo" and method == "pdex":
        # For memory benchmark, we raise here so that it does not show in the resulting summary
        # raise ValueError('PDEX can only run OVO ranksum test')
        pytest.skip("pdex only implements OVO test.")

    # Compile outside of the tracker context
    # if method == "illico":
    #     _precompile(adata.X, reference="non-targeting" if test == "ovo" else None)

    test_params_string = re.match(".*\[(.*)\]", request.node.name).group(1)
    outdir = Path(os.environ.get("MEMRAY_RESULTS_DIR") or Path(__file__).parents[1])

    trace_filepath = lambda x: outdir / ".memray-trackings" / f"trace-{test_params_string}-{str(x).zfill(4)}.bin"
    run_increment = 0
    while (_fp := trace_filepath(run_increment)).exists():
        run_increment += 1
    _fp.parent.mkdir(exist_ok=True)

    try:
        with memray.Tracker(_fp, file_format=memray.FileFormat.AGGREGATED_ALLOCATIONS):
            _ = call_routine(adata, method, test, num_threads)()
    except Exception as e:
        # Cleanup the file if an error happened
        _fp.unlink(missing_ok=True)
        raise e



import numpy as np
from scipy import sparse as sc_sparse
from scipy.stats import rankdata

from illico.utils.ranking import (
    _accumulate_group_ranksums_from_argsort,
    _sort_csc_columns_inplace,
    rank_sum_and_ties_from_sorted,
)
from illico.utils.sparse.csc import CSCMatrix


def test_rank_sum_and_ties_from_sorted():
    rng = np.random.RandomState(0)
    A = rng.randint(0, 10, size=20)
    B = rng.randint(0, 10, size=15)
    A.sort()
    B.sort()

    ranksum_B, tie_sum = rank_sum_and_ties_from_sorted(A, B)

    # Now manually compute ranksum
    combined = np.concatenate([A, B])
    ranks = rankdata(combined, method="average")
    ranksum_B_manual = ranks[len(A) :].sum()
    # Manually compute tie sum
    _, tie_counts = np.unique(combined, return_counts=True)
    manual_tie_sum = (tie_counts**3 - tie_counts).sum()

    # Check
    np.testing.assert_allclose(ranksum_B, ranksum_B_manual)
    np.testing.assert_allclose(tie_sum, manual_tie_sum)


def test_group_ranksum_accumulation():
    rng = np.random.RandomState(0)
    arr = rng.rand(30)
    groups = rng.randint(0, 3, size=30)
    idx = np.argsort(arr)

    ranksums = np.zeros(3, dtype=np.float64)
    tie_sum = _accumulate_group_ranksums_from_argsort(arr, idx, groups, ranksums)

    # Manually compute ranks
    ranks = rankdata(arr, method="average")
    manual_ranksums = np.zeros(3, dtype=np.float64)
    for i in range(len(arr)):
        manual_ranksums[groups[i]] += ranks[i]

    # Manually compute tie sum
    _, tie_counts = np.unique(arr, return_counts=True)
    manual_tie_sum = (tie_counts**3 - tie_counts).sum()

    # Check
    np.testing.assert_allclose(ranksums, manual_ranksums)
    np.testing.assert_allclose(tie_sum, manual_tie_sum)


def test_sort_csc_columns_inplace():

    data = np.array([3, 1, 2, 5, 4], dtype=np.float64)
    indices = np.array([0, 2, 1, 0, 1], dtype=np.int64)
    indptr = np.array([0, 2, 3, 5], dtype=np.int64)  # 3 columns
    csc_matrix = CSCMatrix(data=data, indices=indices, indptr=indptr, shape=(3, 3))

    _sort_csc_columns_inplace(csc_matrix)

    for j in range(csc_matrix.shape[1]):
        col_data = csc_matrix.data[csc_matrix.indptr[j] : csc_matrix.indptr[j + 1]]
        assert np.all(col_data[:-1] <= col_data[1:])



from __future__ import annotations

import os
import time
from typing import Any

from loguru import logger
from numba import types

from illico.utils.groups import GroupContainer
from illico.utils.registry import DataHandler, Test, nb_dispatcher_registry


def _precompile(data_handler: DataHandler, reference: Any | None):
    """Precompile the CPU dispatcher before the threads start rushing to it.

    Note: a simpler way to do it could be to use a threading.lock to make just the first thread compile, but
    there are still concurrency risks.
    Having this in a separate routine also allows to call it from tests so that memory and speed benchmarks are not impacted by compilation.

    Args:

        reference (Any | None): Reference group

    Raises:
        ValueError: If input data is neither dense, CSC nor CSR.

    Author: Rémy Dubois
    """
    if os.environ.get("NUMBA_DISABLE_JIT", "0") == "1":
        logger.warning("Numba JIT is disabled, skipping precompilation.")
        return
    GroupContainerType = types.NamedTuple(
        [types.uint64[::1], types.uint64[::1], types.uint64[::1], types.uint64[::1], types.int64], GroupContainer
    )

    # This input signature corresponds to: lower bound, upper bvound, group container, is_log1p, use_continuity
    common_sig = (
        types.int64,
        types.int64,
        GroupContainerType,
        types.boolean,
        types.boolean,
        types.boolean,
        types.string,
    )
    # This is the output: three float64 2D arrays
    out_sig = types.UniTuple(types.float64[:, ::1], 3)

    input_type = data_handler.input_signature()
    if reference is None:
        test_type = Test.OVR
    else:
        test_type = Test.OVO
    dispatcher = nb_dispatcher_registry.get(test_type, data_handler.kernel_data_format())
    sig = out_sig(input_type, *common_sig)

    s = time.time()
    dispatcher.compile(sig)
    e = time.time()
    logger.trace(
        f"Precompilation of {data_handler.kernel_data_format().value}-{test_type.value} dispatcher took {e - s:.1f}s"
    )
    return dispatcher



import numpy as np
import pytest
from scipy import sparse as sc_sparse

from illico.utils.sparse.csr import (
    _assert_is_csr,
    csr_count_nonzeros,
    csr_get_contig_cols_into_csc,
    csr_get_contig_cols_into_csr,
    csr_get_rows_into_csc,
    csr_to_csc,
)


def test_assert_is_csr(rand_csr):
    _assert_is_csr(rand_csr)  # Should not raise


@pytest.mark.xfail(raises=AssertionError)
def test_assert_is_csr_invalid(rand_csr):
    _assert_is_csr(rand_csr)


@pytest.mark.parametrize("axis", [None, 0, 1])
def test_csr_count_nonzeros(rand_csr, axis):
    nnz = csr_count_nonzeros(rand_csr, axis=axis)
    py_mat = sc_sparse.csr_matrix((rand_csr.data, rand_csr.indices, rand_csr.indptr), shape=rand_csr.shape).toarray()
    expected = np.count_nonzero(py_mat, axis=axis)
    np.testing.assert_allclose(nnz, expected)


def test_csr_get_rows(rand_csr):
    rng = np.random.RandomState(0)
    indices = rng.randint(0, rand_csr.shape[0], size=rand_csr.shape[0] // 2, dtype=np.int32)
    sub_csc = csr_get_rows_into_csc(rand_csr, indices)
    py_sub_csr = sc_sparse.csr_matrix((rand_csr.data, rand_csr.indices, rand_csr.indptr), shape=rand_csr.shape)[
        indices, :
    ].tocsc()
    np.testing.assert_allclose(sub_csc.data, py_sub_csr.data)
    np.testing.assert_allclose(sub_csc.indices, py_sub_csr.indices)
    np.testing.assert_allclose(sub_csc.indptr, py_sub_csr.indptr)
    np.testing.assert_allclose(sub_csc.shape, py_sub_csr.shape)


def test_csr_to_csc(rand_csr):
    csc_matrix = csr_to_csc(rand_csr)
    py_csr = sc_sparse.csr_matrix((rand_csr.data, rand_csr.indices, rand_csr.indptr), shape=rand_csr.shape)
    py_csc = sc_sparse.csc_matrix(py_csr)
    np.testing.assert_allclose(py_csc.data, csc_matrix.data)
    np.testing.assert_allclose(py_csc.indices, csc_matrix.indices)
    np.testing.assert_allclose(py_csc.indptr, csc_matrix.indptr)
    np.testing.assert_allclose(py_csc.shape, csc_matrix.shape)


def test_csr_get_contig_cols_into_csc(rand_csr):
    chunk_lb = rand_csr.shape[1] // 3
    chunk_ub = 2 * rand_csr.shape[1] // 3
    csc_chunk = csr_get_contig_cols_into_csc(rand_csr, chunk_lb, chunk_ub)
    py_csr = sc_sparse.csr_matrix((rand_csr.data, rand_csr.indices, rand_csr.indptr), shape=rand_csr.shape)
    py_chunk = py_csr[:, chunk_lb:chunk_ub].tocsc()
    np.testing.assert_allclose(py_chunk.data, csc_chunk.data)
    np.testing.assert_allclose(py_chunk.indices, csc_chunk.indices)
    np.testing.assert_allclose(py_chunk.indptr, csc_chunk.indptr)
    np.testing.assert_allclose(py_chunk.shape, csc_chunk.shape)


def test_csr_get_contig_cols_into_csr(rand_csr):
    chunk_lb = rand_csr.shape[1] // 3
    chunk_ub = 2 * rand_csr.shape[1] // 3
    csr_chunk = csr_get_contig_cols_into_csr(rand_csr, chunk_lb, chunk_ub)
    py_csr = sc_sparse.csr_matrix((rand_csr.data, rand_csr.indices, rand_csr.indptr), shape=rand_csr.shape)
    py_chunk = py_csr[:, chunk_lb:chunk_ub]
    np.testing.assert_allclose(py_chunk.data, csr_chunk.data)
    np.testing.assert_allclose(py_chunk.indices, csr_chunk.indices)
    np.testing.assert_allclose(py_chunk.indptr, csr_chunk.indptr)
    np.testing.assert_allclose(py_chunk.shape, csr_chunk.shape)



from __future__ import annotations

# from abc import ABC, abstractmethod
from collections import namedtuple

# import anndata as ad
# import h5py
# import numpy as np
# from scipy import sparse

TestResults = namedtuple("TestResults", ["statistic", "pvalue"])
# CSRMatrix = namedtuple("CSRMatrix", ["data", "indices", "indptr", "shape"])
# CSCMatrix = namedtuple("CSCMatrix", ["data", "indices", "indptr", "shape"])


# def scipy_to_nb(
#     obj: np.ndarray | sparse.csc_matrix | sparse.csr_matrix,
# ) -> np.ndarray | CSRMatrix | CSCMatrix:
#     """Convert python object into Numba-compat holder

#     Returns
#     -------
#     np.ndarray | CSRMatrix | CSCMatrix
#         The Numba-compatible object

#     Raises
#     ------
#     ValueError
#         If input is neither dense, csc or csr.

#     Author: Rémy Dubois
#     """
#     if isinstance(obj, np.ndarray):
#         return obj
#     elif isinstance(obj, sparse.csc_matrix):
#         return CSCMatrix(obj.data, obj.indices, obj.indptr, obj.shape)
#     elif isinstance(obj, sparse.csr_matrix):
#         return CSRMatrix(obj.data, obj.indices, obj.indptr, obj.shape)
#     else:
#         raise ValueError(type(obj))

# def fetch_if_backed(f):
#     """
#     Decorator to fetch chunks of backed dataset in RAM for backed mode.
#     Here, f is either:
#     - ovo_mwu_over_contiguous_col_chunk
#     - ovr_mwu_over_contiguous_col_chunk
#     """

#     def g(
#         X,
#         chunk_lb,
#         chunk_ub,
#         group_container,
#         is_log1p,
#         use_continuity,
#         alternative):
#         if isinstance(X, h5py.Dataset):
#             X = X[:, chunk_lb:chunk_ub]
#             chunk_ub = chunk_ub - chunk_lb
#             chunk_lb = 0
#         elif isinstance(X, ad._core.sparse_dataset._CSCDataset):
#             X = X[:, chunk_lb:chunk_ub]
#             chunk_ub = chunk_ub - chunk_lb
#             chunk_lb = 0
#         elif isinstance(X, ad._core.sparse_dataset._CSRDataset):
#             raise ValueError("CSR backed AnnData is not supported.")
#         return f(
#             X=X,
#             chunk_lb=chunk_lb,
#             chunk_ub=chunk_ub,
#             group_container=group_container,
#             is_log1p=is_log1p,
#             use_continuity=use_continuity,
#             alternative=alternative)

#     return g



from abc import ABC, abstractmethod
from enum import Enum
from typing import Any

import anndata as ad
import h5py
import numpy as np
from numba import types
from scipy import sparse as py_sparse

from illico.utils.sparse.csc import CSCMatrix
from illico.utils.sparse.csr import CSRMatrix


class Test(Enum):
    OVO = "ovo"
    OVR = "ovr"


class KernelDataFormat(Enum):
    DENSE = "dense"
    CSC = "csc"
    CSR = "csr"


class DispatcherRegistry(dict):
    def register(self, test: Test, data_format: KernelDataFormat):
        test = Test(test)
        data_format = KernelDataFormat(data_format)

        def decorator(obj):
            key = (test, data_format)
            self[key] = obj
            return obj

        return decorator

    def get(self, test: Test, data_format: KernelDataFormat):
        key = (Test(test), KernelDataFormat(data_format))
        try:
            return self[key]
        except KeyError as e:
            raise KeyError(f"No dispatcher registered for test {test} and data format {data_format}.") from e


class DataHandlerRegistry(dict):
    def register(self, data_format):
        def decorator(obj):
            self[data_format] = obj
            return obj

        return decorator

    def get(self, key):
        try:
            return self[type(key)](key)
        except KeyError as e:
            raise KeyError(f"Support for data type {type(key)} is not implemented.") from e


# How to fetch data from disk, if data is backed or lazy-loaded
data_handler_registry = DataHandlerRegistry()
# Which dispatcher to use depending on data format and test type
nb_dispatcher_registry = DispatcherRegistry()
# Register the same dispatchers for the Rust kernels
rs_dispatcher_registry = DispatcherRegistry()


class DataHandler(ABC):
    def __init__(self, data):
        self.data = data

    @abstractmethod
    def input_signature(self, *args, **kwargs) -> tuple:
        """Return the numba input signature for this handler."""
        pass

    @abstractmethod
    def fetch(self, *args, **kwargs) -> tuple:
        """Fetch data from disk if needed."""
        pass

    @abstractmethod
    def to_nb(self, *args, **kwargs) -> Any:
        """Convert data to numba-compatible format."""
        pass

    @abstractmethod
    def kernel_data_format(self) -> KernelDataFormat:
        """Return the dispatcher kernel routine for this handler."""
        pass

    @abstractmethod
    def footprint(self) -> int:
        """Return estimated memory footprint of the data."""
        pass


class InRAMDataHandler(DataHandler):
    def fetch(self, lb: int, ub: int) -> tuple:
        """If the data is already in RAM, let the kernels do optimized slicing."""
        return self.data, (lb, ub)


@data_handler_registry.register(np.ndarray)
class DenseDataHandler(InRAMDataHandler):
    def input_signature(self) -> tuple:
        # Because this will be loaded by chunk, input type is necessarily contiguous
        input_type = getattr(types, str(self.data.dtype))[:, ::1]
        return input_type

    def kernel_data_format(self) -> KernelDataFormat:
        return KernelDataFormat.DENSE

    def footprint(self) -> int:
        return self.data.nbytes

    @classmethod
    def to_nb(cls, X: np.ndarray) -> np.ndarray:
        assert isinstance(X, np.ndarray)
        return X


@data_handler_registry.register(py_sparse._csr.csr_matrix)
class CSRDataHandler(InRAMDataHandler):
    def input_signature(self) -> tuple:
        data_type = getattr(types, str(self.data.data.dtype))[::1]
        indices_type = getattr(types, str(self.data.indices.dtype))[::1]
        indptr_type = getattr(types, str(self.data.indptr.dtype))[::1]
        return types.NamedTuple([data_type, indices_type, indptr_type, types.UniTuple(types.int64, 2)], CSRMatrix)

    @classmethod
    def to_nb(cls, X: py_sparse.csr_matrix) -> CSRMatrix:
        assert isinstance(X, py_sparse.csr.csr_matrix)
        return CSRMatrix(X.data, X.indices, X.indptr, X.shape)

    def kernel_data_format(self) -> KernelDataFormat:
        return KernelDataFormat.CSR

    def footprint(self) -> int:
        return self.data.data.nbytes + self.data.indptr.nbytes + self.data.indices.nbytes


@data_handler_registry.register(py_sparse._csc.csc_matrix)
class CSCDataHandler(InRAMDataHandler):
    def input_signature(self) -> tuple:
        data_type = getattr(types, str(self.data.data.dtype))[::1]
        indices_type = getattr(types, str(self.data.indices.dtype))[::1]
        indptr_type = getattr(types, str(self.data.indptr.dtype))[::1]
        return types.NamedTuple([data_type, indices_type, indptr_type, types.UniTuple(types.int64, 2)], CSCMatrix)

    @classmethod
    def to_nb(cls, X: py_sparse.csc_matrix) -> CSCMatrix:
        assert isinstance(X, py_sparse.csc.csc_matrix)
        return CSCMatrix(X.data, X.indices, X.indptr, X.shape)

    def kernel_data_format(self) -> KernelDataFormat:
        return KernelDataFormat.CSC

    def footprint(self) -> int:
        return self.data.data.nbytes + self.data.indptr.nbytes + self.data.indices.nbytes


@data_handler_registry.register(h5py.Dataset)
class H5pyDatasetDataHandler(DenseDataHandler):
    def fetch(self, lb: int, ub: int) -> tuple:
        return self.data[:, lb:ub], (0, ub - lb)

    def footprint(self):
        return self.data.nbytes


@data_handler_registry.register(ad._core.sparse_dataset._CSCDataset)
class H5pyBackedCSCDataHandler(CSCDataHandler):
    def input_signature(self) -> tuple:
        data_type = getattr(types, str(self.data._data.dtype))[::1]
        indices_type = getattr(types, str(self.data._indices.dtype))[::1]
        indptr_type = getattr(types, str(self.data._indptr.dtype))[::1]
        return types.NamedTuple([data_type, indices_type, indptr_type, types.UniTuple(types.int64, 2)], CSCMatrix)

    @classmethod
    def to_nb(cls, X: py_sparse.csc_matrix) -> CSCMatrix:
        assert isinstance(X, py_sparse.csc.csc_matrix)
        return CSCMatrix(X.data, X.indices, X.indptr, X.shape)

    def footprint(self) -> int:
        return self.data._data.nbytes + self.data._indptr.nbytes + self.data._indices.nbytes

    def fetch(self, lb: int, ub: int) -> tuple:
        return self.data[:, lb:ub], (0, ub - lb)


# Import kernel modules to trigger decorator registration
# These imports must come after the registry definitions above
from illico.ovo import (  # noqa: E402, F401
    csc_ovo_mwu_kernel_over_contiguous_col_chunk,
    csr_ovo_mwu_kernel_over_contiguous_col_chunk,
    dense_ovo_mwu_kernel_over_contiguous_col_chunk,
)
from illico.ovr import (  # noqa: E402, F401
    csc_ovr_mwu_kernel_over_contiguous_col_chunk,
    csr_ovr_mwu_kernel_over_contiguous_col_chunk,
    dense_ovr_mwu_kernel_over_contiguous_col_chunk,
)

# Now register the Rust kernels
from illico.rust_backend import (  # noqa: E402, F401
    csc_ovo_mwu_kernel_over_contiguous_col_chunk_rust,
    csc_ovr_mwu_kernel_over_contiguous_col_chunk_rust,
    csr_ovo_mwu_kernel_over_contiguous_col_chunk_rust,
    csr_ovr_mwu_kernel_over_contiguous_col_chunk_rust,
    dense_ovo_over_contiguous_col_chunk_rust,
    dense_ovr_over_contiguous_col_chunk_rust,
)

rs_dispatcher_registry.register(Test.OVO, KernelDataFormat.DENSE)(dense_ovo_over_contiguous_col_chunk_rust)
rs_dispatcher_registry.register(Test.OVR, KernelDataFormat.DENSE)(dense_ovr_over_contiguous_col_chunk_rust)
rs_dispatcher_registry.register(Test.OVO, KernelDataFormat.CSC)(csc_ovo_mwu_kernel_over_contiguous_col_chunk_rust)
rs_dispatcher_registry.register(Test.OVO, KernelDataFormat.CSR)(csr_ovo_mwu_kernel_over_contiguous_col_chunk_rust)
rs_dispatcher_registry.register(Test.OVR, KernelDataFormat.CSC)(csc_ovr_mwu_kernel_over_contiguous_col_chunk_rust)
rs_dispatcher_registry.register(Test.OVR, KernelDataFormat.CSR)(csr_ovr_mwu_kernel_over_contiguous_col_chunk_rust)



import numpy as np
import pytest
from scipy import sparse as sc_sparse

from illico.utils.sparse.csc import (
    _assert_is_csc,
    csc_count_nonzeros,
    csc_get_cols,
    csc_get_contig_cols_into_csr,
    csc_to_csr,
)


def test_assert_is_csc(rand_csc):
    _assert_is_csc(rand_csc)  # Should not raise


@pytest.mark.xfail(raises=AssertionError)
def test_assert_is_csc_invalid(rand_csc):
    _assert_is_csc(rand_csc)


@pytest.mark.parametrize("axis", [None, 0, 1])
def test_csc_count_nonzeros(rand_csc, axis):
    nnz = csc_count_nonzeros(rand_csc, axis=axis)
    py_mat = sc_sparse.csc_matrix((rand_csc.data, rand_csc.indices, rand_csc.indptr), shape=rand_csc.shape).toarray()
    expected = np.count_nonzero(py_mat, axis=axis)
    np.testing.assert_allclose(nnz, expected)


def test_csc_get_cols(rand_csc):
    rng = np.random.RandomState(0)
    indices = rng.randint(0, rand_csc.shape[1], size=rand_csc.shape[1] // 2, dtype=np.int32)
    sub_csc = csc_get_cols(rand_csc, indices)
    py_sub_csc = sc_sparse.csc_matrix((rand_csc.data, rand_csc.indices, rand_csc.indptr), shape=rand_csc.shape)[
        :, indices
    ]
    np.testing.assert_allclose(sub_csc.data, py_sub_csc.data)
    np.testing.assert_allclose(sub_csc.indices, py_sub_csc.indices)
    np.testing.assert_allclose(sub_csc.indptr, py_sub_csc.indptr)
    np.testing.assert_allclose(sub_csc.shape, py_sub_csc.shape)


def test_csc_to_csr(rand_csc):
    csr_matrix = csc_to_csr(rand_csc)
    py_csc = sc_sparse.csc_matrix((rand_csc.data, rand_csc.indices, rand_csc.indptr), shape=rand_csc.shape)
    py_csr = sc_sparse.csr_matrix(py_csc)
    np.testing.assert_allclose(py_csr.data, csr_matrix.data)
    np.testing.assert_allclose(py_csr.indices, csr_matrix.indices)
    np.testing.assert_allclose(py_csr.indptr, csr_matrix.indptr)
    np.testing.assert_allclose(py_csr.shape, csr_matrix.shape)


def test_csc_get_contig_cols_into_csr(rand_csc):
    chunk_lb = rand_csc.shape[1] // 3
    chunk_ub = 2 * rand_csc.shape[1] // 3
    csr_chunk = csc_get_contig_cols_into_csr(rand_csc, chunk_lb, chunk_ub)
    py_csc = sc_sparse.csc_matrix((rand_csc.data, rand_csc.indices, rand_csc.indptr), shape=rand_csc.shape)
    py_chunk = py_csc[:, chunk_lb:chunk_ub].tocsr()
    np.testing.assert_allclose(py_chunk.data, csr_chunk.data)
    np.testing.assert_allclose(py_chunk.indices, csr_chunk.indices)
    np.testing.assert_allclose(py_chunk.indptr, csr_chunk.indptr)
    np.testing.assert_allclose(py_chunk.shape, csr_chunk.shape)


# def test_csc_to_dense(rand_csc):
#     dense_mat = csc_to_dense(rand_csc)
#     py_csc = sc_sparse.csc_matrix(
#         (rand_csc.data, rand_csc.indices, rand_csc.indptr), shape=rand_csc.shape
#     )
#     py_dense = py_csc.toarray()
#     np.testing.assert_allclose(py_dense, dense_mat)



from illico.ovo.dense_ovo import (
    dense_ovo_mwu_kernel,
    dense_ovo_mwu_kernel_over_contiguous_col_chunk,
)
from illico.ovo.sparse_ovo import (
    csc_ovo_mwu_kernel_over_contiguous_col_chunk,
    csr_ovo_mwu_kernel_over_contiguous_col_chunk,
)

__all__ = [
    "dense_ovo_mwu_kernel",
    "dense_ovo_mwu_kernel_over_contiguous_col_chunk",
    "csc_ovo_mwu_kernel_over_contiguous_col_chunk",
    "csr_ovo_mwu_kernel_over_contiguous_col_chunk",
]



import math
from typing import Literal

import anndata as ad
import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from loguru import logger
from scipy import sparse
from tqdm.auto import tqdm

from illico.utils.compile import _precompile
from illico.utils.groups import GroupContainer, encode_and_count_groups
from illico.utils.memory import log_memory_usage
from illico.utils.ranking import check_indices_sorted_per_parcel
from illico.utils.registry import (
    DataHandler,
    Test,
    data_handler_registry,
    nb_dispatcher_registry,
    rs_dispatcher_registry,
)

# from illico.utils.math import _warn_log1p

__all__ = ["asymptotic_wilcoxon"]


@delayed
def operator(
    data_handler: DataHandler,
    lb: int,
    ub: int,
    group_container: GroupContainer,
    is_log1p: bool,
    use_continuity: bool,
    alternative: str,
    tie_correct: bool,
    use_rust: bool,
    results: np.ndarray,
):
    """Delayed operator. Not user-facing."""
    if group_container.encoded_ref_group == -1:
        test = Test.OVR
    else:
        test = Test.OVO
    # Grab the adapted kernel
    if not use_rust:
        dispatcher = nb_dispatcher_registry.get(test, data_handler.kernel_data_format())
    else:
        dispatcher = rs_dispatcher_registry.get(test, data_handler.kernel_data_format())

    # Quick safety check
    if lb < 0 or ub > data_handler.data.shape[1] or lb > ub:
        raise ValueError(f"Invalid chunk bounds: {(lb, ub)} for data with {data_handler.data.shape[1]} columns.")

    # Fetch the data from disk if in backed mode
    # The reason to be of not applying X[:, lb:ub] in all cases (backed or not) is that if the data is whole in RAM, the CSR chunking is optimized, and
    # if the data is not in RAM, CSR chunking is not implemented
    fetched_data, bounds = data_handler.fetch(lb, ub)
    # Convert to numba-compatible format
    X = data_handler.to_nb(fetched_data)
    # Call the dispatcher
    pvalues, statistics, fold_change = dispatcher(
        X,
        *bounds,
        group_container,
        is_log1p,
        use_continuity,
        tie_correct,
        alternative,
    )
    # Copy results into the shared array, do it thread-wise for cleaner garbage collection and speed
    # Note: there might be a little speedup in passing results to the dispatchers and writing in it directly, it would
    # allow to not allocate chunks of results. However, this is would be very non-Rusty as the same Python-managed memory block would be shared across several threads.
    # The copy should be GIL-free
    results[:, lb:ub, 0] = pvalues
    results[:, lb:ub, 1] = statistics
    results[:, lb:ub, 2] = fold_change  # Technically Rust returns f32, but numpy handles casting here
    return (lb, ub)


def asymptotic_wilcoxon(
    adata: ad.AnnData,
    is_log1p: bool,
    group_keys: str,
    reference: str | None = None,
    n_threads: int = 1,
    batch_size: int | Literal["auto"] = "auto",
    alternative: str = "two-sided",
    use_continuity: bool = True,
    tie_correct: bool = True,
    layer: str | None = None,
    precompile: bool = True,
    use_rust: bool = True,
):
    """Perform asymptotic Mann-Whitney tests for differential gene expression.

    Mann-Whitney test is the same as Wilcoxon rank-sum test.
    This function takes as input an AnnData object of shape (n_cells, n_genes) with a group
    (e.g., perturbation) variable stored in .obs. It performs either one-versus-rest (OVR) or
    one-versus-one (OVO) Wilcoxon-Mann-Whitney tests for each gene, depending on whether a
    reference group is provided.

    It supports in-RAM dense, sparse CSC and sparse CSR matrices, as well as backed dense and sparse CSC matrices.

    Parameters
    ----------
    adata : ad.AnnData
        Annotated data matrix of shape (n_cells, n_genes).
    is_log1p : bool
        Whether the data is log1p transformed.
    group_keys : str
        Key in `adata.obs` specifying the group variable.
    reference : str or None, default=None
        Name of the reference group for OVO tests. If `None`, OVR tests are performed.
    n_threads : int, default=1
        Number of threads to use for parallel computation.
    batch_size : int or "auto", default="auto"
        Number of genes to process in each batch. If "auto", automatically determines
        optimal batch size aiming for approximately 256 genes per chunk.
    alternative : str, default="two-sided"
        Type of alternative hypothesis. One of 'two-sided', 'less', or 'greater'.
    use_continuity : bool, default=True
        Whether to apply continuity correction.
    tie_correct : bool, default=True
        Whether to apply tie correction in the test statistic.
    layer : str or None, default=None
        Layer in `adata.layers` to use for the data. If `None`, uses `adata.X`.
    precompile : bool, default=True
        Whether to precompile necessary functions for performance. It is recommended to set this to `True`.
    use_rust : bool, default=True
        Whether to use the Rust implementation of the test. If `False`, uses the Numba implementation.

    Returns
    -------
    pd.DataFrame
        A DataFrame with MultiIndex (pert, feature) containing columns:
        - 'p_value': P-value from the Mann-Whitney test
        - 'statistic': Test statistic (U-statistic)
        - 'fold_change': Fold change between groups

    Raises
    ------
    ValueError
        If input data matrix indices are not sorted (for sparse CSR matrices).
        If batch_size is not 'auto' or an integer.

    Examples
    --------
    >>> import anndata as ad
    >>> import numpy as np
    >>> import pandas as pd
    >>>
    >>> # Create example AnnData object
    >>> n_cells, n_genes = 1000, 500
    >>> X = np.random.negative_binomial(5, 0.3, size=(n_cells, n_genes))
    >>> obs = pd.DataFrame({'cell_type': np.random.choice(['A', 'B', 'C'], n_cells)})
    >>> var = pd.DataFrame(index=[f'gene_{i}' for i in range(n_genes)])
    >>> adata = ad.AnnData(X=X, obs=obs, var=var)
    >>>
    >>> # Perform one-versus-rest tests
    >>> results = asymptotic_wilcoxon(
    ...     adata,
    ...     is_log1p=False,
    ...     group_keys='cell_type',
    ...     n_threads=4
    ... )
    >>> print(results.head())
    >>>
    >>> # Perform one-versus-one tests against reference
    >>> results_ovo = asymptotic_wilcoxon(
    ...     adata,
    ...     is_log1p=False,
    ...     group_keys='cell_type',
    ...     reference='A',
    ...     n_threads=4,
    ...     alternative='greater'
    ... )
    >>>
    >>> # Filter significant results
    >>> significant = results[results['p_value'] < 0.05]
    >>> print(f"Found {len(significant)} significant tests")

    Notes
    -----
    The function automatically handles both dense and sparse matrices. For sparse CSR matrices,
    indices must be sorted per row to ensure correct results.

    Author: Rémy Dubois
    """

    # Get expression matrix
    if layer is not None:
        logger.info(f"Using layer '{layer}' for differential expression.")
        X = adata.layers[layer]
    else:
        X = adata.X
    data_handler = data_handler_registry.get(X)

    # Check that the input CSR is sorted.
    if isinstance(X, sparse.csr_matrix):
        if not check_indices_sorted_per_parcel(X.indices, X.indptr):
            raise ValueError(
                "Input data matrix indices are not sorted. This is very unusual and may lead to incorrect results. "
                "This can be the result of operations like `adata[:, np.random.choice(…)]` that do not preserve sorting."
                "Please make sure that indices used to chunk the adata or the expression matrix have been sorted "
                "prior to computing DE genes."
            )

    # Precompile if requested
    if precompile:
        if use_rust:
            logger.info("No precompilation needed for Rust kernels.")
        else:
            _precompile(data_handler, reference)

    # Process the groups information
    unique_raw_groups, group_container = encode_and_count_groups(
        groups=adata.obs[group_keys].values, ref_group=reference
    )
    logger.info(
        f"Found {group_container.counts.size} unique groups (min size: {group_container.counts.min()} cells; max size: {group_container.counts.max()} cells), with reference group: {reference}"
    )
    _, n_genes = X.shape

    # Allocate the results dataframes
    cols = pd.Series(adata.var_names, name="feature", dtype=str)
    rows = pd.Series(unique_raw_groups, name="pert", dtype=str)
    results = np.empty((len(rows), len(cols), 3), dtype=np.float64)

    # Adapt batch size to leverage multithreading regarding the number of genes, if requested
    if n_genes < 256:
        batch_size = n_genes  # No batching for small number of genes
        n_threads = 1  # No multithreading for small number of genes
        iterator = [[0, n_genes]]
    elif isinstance(batch_size, int):
        batch_size = min(batch_size, math.ceil(n_genes / n_threads))
        bounds = np.append(np.arange(0, n_genes, batch_size), n_genes)
        iterator = list(zip(bounds[:-1], bounds[1:]))
    elif batch_size == "auto":
        n_dispatches = max(int(n_genes / 256 / n_threads), 1)  # Aim for approximately 256 genes per chunk
        splits = np.array_split(np.arange(n_genes + 1), indices_or_sections=n_threads * n_dispatches)
        iterator = [split[[0, -1]] for split in splits]
        batch_size = int(np.ceil(n_genes / (n_dispatches * n_threads)))
    else:
        raise ValueError(f"Invalid batch_size value: {batch_size}. Must be 'auto' or an integer.")
    logger.trace(f"Using batch size of {batch_size} for {n_threads} threads and {n_genes} genes.")

    # Compute estimated mem footprint
    _ = log_memory_usage(data_handler, group_container, batch_size, n_threads)

    # Go through all the possible combinations
    n_tests = n_genes * group_container.counts.size
    logger.trace(f"Performing a total of {n_tests:,d} tests.")
    with Parallel(n_threads, prefer="threads", return_as="generator_unordered") as pool:
        with tqdm(total=n_tests, smoothing=0.0, unit="it", unit_scale=True, unit_divisor=1000) as pbar:
            for lb, ub in pool(
                operator(
                    data_handler,
                    lb,
                    ub,
                    group_container,
                    is_log1p,
                    use_continuity,
                    alternative,
                    tie_correct,
                    use_rust,
                    results,
                )
                for lb, ub in iterator
            ):
                pbar.update(group_container.counts.size * (ub - lb))

        # Return a pd.DataFrame to index results
        results = pd.DataFrame(
            data=results.reshape(-1, 3),
            index=pd.MultiIndex.from_product([rows, cols], names=["pert", "feature"]),
            columns=["p_value", "statistic", "fold_change"],
        )

    return results



import anndata as ad
import pytest

from illico.utils.compile import _precompile
from illico.utils.groups import encode_and_count_groups
from illico.utils.registry import data_handler_registry


@pytest.mark.parametrize("test", ["ovo", "ovr"])
def test_precompile(rand_adata, test):
    # No need to test that exception is raised, as it is done in `test_asymptotic_wilcoxon` already
    if isinstance(rand_adata.X, ad._core.sparse_dataset._CSRDataset):
        pytest.skip("CSR lazy data not supported for now.")

    reference = rand_adata.obs.pert.iloc[0] if test == "ovo" else None
    # Now compile it, and make sure it compiled nopython
    data_handler = data_handler_registry.get(rand_adata.X)
    dispatcher = _precompile(data_handler, reference)
    assert (len(leg_sig := dispatcher.nopython_signatures)) > 0, "Dispatcher should be compiled now."
    # Now run the dispatcher
    _, grpc = encode_and_count_groups(rand_adata.obs.pert.values, reference)
    X, bounds = data_handler.fetch(0, rand_adata.X.shape[1])
    X_nb = data_handler.to_nb(X)
    dispatcher(X_nb, *bounds, grpc, False, True, True, "two-sided")
    # Assert no other signature was added
    assert len(dispatcher.nopython_signatures) == len(
        leg_sig
    ), f"Dispatcher should not have recompiled: {chr(10).join(map(str, dispatcher.nopython_signatures))}"



import numpy as np
import pytest

from illico.utils.math import _add_at_scalar, _add_at_vec, _warn_log1p, diff


def test_add_at_scalar():
    a = np.zeros(5, dtype=np.float64)
    b = np.array([0, 2, 4, 2, 3, 0], dtype=np.int64)
    c = 1.0
    _add_at_scalar(a, b, c)
    expected = np.zeros(5, dtype=np.float64)
    np.add.at(expected, b, c)
    np.testing.assert_allclose(a, expected)


def test_add_at_vec():
    a = np.zeros(5, dtype=np.float64)
    b = np.array([0, 2, 4, 2, 3, 0], dtype=np.int64)
    c = np.array([1.0, 2.0, 3.0, 4.0, 5.0, 6.0], dtype=np.float64)
    _add_at_vec(a, b, c)
    expected = np.zeros(5, dtype=np.float64)
    np.add.at(expected, b, c)
    np.testing.assert_allclose(a, expected)


def test_diff():
    x = np.array([1, 3, 6, 10, 15], dtype=np.int64)
    result = diff(x)
    expected = np.diff(x)
    np.testing.assert_array_equal(result, expected)


""" --- IGNORE --- as warn_log1p is no longer used."""
# def test_warn_log1p_1(rand_adata, data_format):
#     # Case 1: Data is raw counts, is_log1p=True
#     with pytest.warns(UserWarning):
#         _warn_log1p(rand_adata.layers[data_format], is_log1p=True)


# def test_warn_log1p_2(rand_adata, data_format):
#     # Case 2: Data is raw counts, is_log1p=False (no warning)
#     _warn_log1p(rand_adata.layers[data_format], is_log1p=False)


# def test_warn_log1p_3(rand_adata, data_format):
#     # Case 3: Data is log1p transformed, is_log1p=False
#     rand_adata.layers[data_format] = np.log1p(rand_adata.layers[data_format])
#     with pytest.warns(UserWarning):
#         _warn_log1p(rand_adata.layers[data_format], is_log1p=False)


# def test_warn_log1p_4(rand_adata, data_format):
#     # Case 4: Data is log1p transformed, is_log1p=True (no warning)
#     _warn_log1p(rand_adata.layers[data_format], is_log1p=True)
